import datetime
import contextlib
import requests
from freezegun import freeze_time
from unittest.mock import patch

import odoo
from odoo import fields, Command, SUPERUSER_ID
from odoo.exceptions import UserError, ValidationError
from odoo.tests import HttpCase, Form, tagged, users, get_db_name, new_test_user
from odoo.tools import mute_logger
from odoo.addons.account.tests.common import AccountTestInvoicingCommon
from odoo.addons.account_payment.tests.common import AccountPaymentCommon
from odoo.addons.sale_subscription.tests import test_payments as sale_subscription_test_payments
from odoo.addons.openerp_enterprise.tests.common import (
    EnterpriseDatabaseCommon, generate_uuid, now, FakeDatetime, monkey_patch
)
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.addons.saas_worker.tests.common import SaasWorkerHttpCommon


# since 16.0, *PaymentCommon classes does not propagate the chart_template_ref,
# so allow specifying it using class attribute instead
AccountTestInvoicingCommon.chart_template_ref = None

@monkey_patch(AccountTestInvoicingCommon, 'setUpClass')
@classmethod
def AccountTestInvoicingCommon_setUpClass(cls, chart_template_ref=None):
    chart_template_ref = chart_template_ref or cls.chart_template_ref

    # protect main company against currency change - if company has a non-USD
    # currency and is loaded with demo data it will crash; accounting tests
    # should anyway create and test with their own company(ies)
    registry = odoo.registry(get_db_name())
    orig_ResCompany_write = registry['res.company'].write

    def ResCompany_write(self, vals):
        if len(self) == 1 and self.id == 1:
            # skip write of 'currency_id' for base.main_company
            vals = {f: v for f, v in vals.items() if f != 'currency_id'}
        return orig_ResCompany_write(self, vals)

    protect_main_company_currency = patch.object(
        registry['res.company'], 'write', side_effect=ResCompany_write, autospec=True
    )

    with protect_main_company_currency:
        AccountTestInvoicingCommon_setUpClass.super(cls, chart_template_ref=chart_template_ref)


@tagged('post_install', '-at_install')
class TestSubscriptionPayments(AccountPaymentCommon, sale_subscription_test_payments.TestSubscriptionPayments):
    allow_inherited_tests_method = True

    @classmethod
    def setUpClass(cls, chart_template_ref=None):
        # force l10n_be COA template as `openerp_enterprise` depends on `l10n_be`
        # and will be used for Odoo S.A.
        cls.chart_template_ref = "be_comp"
        super().setUpClass()

        context_no_mail = {'no_reset_password': True, 'mail_create_nosubscribe': True, 'mail_create_nolog': True}
        # Analytic = cls.env['account.analytic.account'].with_context(context_no_mail)
        # SaleOrder = cls.env['sale.order'].with_context(context_no_mail)
        Tax = cls.env['account.tax'].with_context(context_no_mail)

        # TestSubscriptionCommon will force company country to 'base.us' we
        # need to override it to match the COA country
        cls.company.country_id = cls.env.ref('base.be')

        # Setup 2nd company
        cls.company_2 = cls.company_data_2["company"]
        cls.company_2_tax_10 = Tax.create({
            'company_id': cls.company_2.id,
            'name': "10% tax",
            'amount_type': 'percent',
            'amount': 10,
        })
        cls.company_2_tax_0 = Tax.create({
            'company_id': cls.company_2.id,
            'name': "0% extra-comm",
            'amount_type': 'percent',
            'amount': 0,
        })
        (cls.product | cls.product2 | cls.product3 | cls.product4 | cls.product5).write({
            'taxes_id': [(4, cls.company_2_tax_10.id)]
        })
        cls.company_2_extracomm_fpos = cls.env["account.fiscal.position"].search([
            ('company_id', '=', cls.company_2.id),
            ('name', 'ilike', 'Import/Export'),
        ])
        cls.company_2_extracomm_fpos.write({
            'tax_ids': [
                Command.create({'tax_src_id': cls.company_2_tax_10.id, 'tax_dest_id': cls.company_2_tax_0.id})
            ]
        })
        cls.acquirer_2 = cls._prepare_provider(company=cls.company_2)

        # TestSubscriptionPayments will call `_cron_recurring_create_invoice()`
        # as "accountant" resulting in AccessError (when accessing the payment
        # token acquirer company). In real life, that method will only be called
        # by a cron which will have appropriate rights (or as superuser).
        @monkey_patch(cls.env.registry["sale.order"], "_cron_recurring_create_invoice")
        def _patched_cron_recurring_create_invoice(self, *args, **kwargs):
            superuser = self.env.ref('base.user_root')
            self = self.with_user(superuser)
            if not self:
                sub_ids = self._recurring_invoice_get_ids(company_id=cls.company | cls.company_2)
                self = self.browse(sub_ids)
            _patched_cron_recurring_create_invoice.super(self, *args, **kwargs)
        cls.addClassCleanup(_patched_cron_recurring_create_invoice.reset)

        # bypass subscription's copy restriction when duplicating the template
        cls.subscription_tmpl = cls.subscription_tmpl.with_context(force_copy_subscription=True)
        cls.subscription_tmpl.company_id = False  # allow to be used in multiple companies
        cls.subscription = cls.subscription.with_context(force_copy_subscription=True)
        cls.env = cls.env(context=dict(cls.env.context, force_copy_subscription=True))

    def _patch_model_default(self, model_name, field_name, return_value):
        return patch.object(self.env[model_name]._fields[field_name], 'default', return_value=return_value)

    @contextlib.contextmanager
    def _mock_enable_order_payment_with_token(self):
        # add default payment token is none provided and enable mock for testing payment w/ token
        with self._patch_model_default('sale.order', 'payment_token_id', self.payment_token), \
            patch('odoo.addons.sale_subscription.models.sale_order.SaleOrder._do_payment',
                    wraps=self._mock_subscription_do_payment),\
            patch('odoo.addons.sale_subscription.models.sale_order.SaleOrder._send_success_mail',
                    wraps=self._mock_subscription_send_success_mail):
            yield

    def test_refund_next_invoice_date(self):
        # std. test create subscription without quotation template nor payment token,
        # force adding them and mock for automatic payment with token
        patch_model_default = self._patch_model_default

        quotation_template_monthly = self.env.ref('openerp_enterprise.custom_template_odoo_monthly')
        quotation_template_monthly.plan_id = self.plan_month

        with self._mock_enable_order_payment_with_token(), \
                patch_model_default('sale.order', 'sale_order_template_id', quotation_template_monthly):
            super().test_refund_next_invoice_date()

    def _mock_subscription_do_payment_and_reconcile(self, payment_token, invoice, auto_commit=False):
        tx_obj = self.env['payment.transaction']
        values = {
            'provider_id': payment_token.provider_id.id,
            'sale_order_ids': [Command.link(self.subscription.id)],
            'amount': invoice.amount_total,
            'currency_id': invoice.currency_id.id,
            'partner_id': self.subscription.partner_id.id,
            'payment_method_id': payment_token.payment_method_id.id,
            'token_id': payment_token.id,
            'operation': 'offline',
            'invoice_ids': [(6, 0, [invoice.id])],
            'subscription_action': 'automatic_send_mail',
        }
        tx = tx_obj.create(values)
        tx._set_done()
        return tx

    @mute_logger('odoo.addons.sale_subscription.models.sale_order')
    def test_exception_mail(self):
        self.subscription.write({
            'payment_token_id': self.payment_token.id,
        })
        self.subscription.action_confirm()
        self.subscription.client_order_ref = "MCUSTREF4TEST"
        SaleOrder = self.env["sale.order"]
        with patch.object(type(SaleOrder), "_do_payment", side_effect=Exception("Bad Token")), \
             self.mock_mail_gateway():
            self.subscription._create_recurring_invoice()
        found_mail = self._find_mail_mail_wemail('online@openerp.com', 'sent', author=self.env.user.partner_id)
        self.assertIn(
            "Error during renewal of contract [%s] MCUSTREF4TEST" % self.subscription.id, found_mail.body_html,
            "Contract Number must be present in the exception email",
        )
        self.assertIn(
            "Bad Token", found_mail.body_html,
            "Failure reason must be present in the exception email",
        )

    def test_manual_create_invoices_access_rights(self):
        """When salesperson manually click on "Create Invcoies", check it's access to the subscription invocing company"""

        # drop administration privileges from accountman
        self.env.user.groups_id -= (
            self.env.ref("base.group_erp_manager") | self.env.ref("base.group_system")
        )
        self.assertEqual(self.subscription.company_id, self.company)
        self.assertFalse(self.env.su)
        self.subscription.action_confirm()

        # Restrict accountant access to only "company_2"
        self.env.user.write({'company_id': self.company_2.id, 'company_ids': self.company_2.ids})
        with self.assertRaisesRegex(UserError, "you don't have access to that company"):
            subscription = self.subscription
            subscription._create_invoices()

        # Set accountant to both companies, but only enable "company_2" in context
        self.env.user.write({'company_id': self.company_2.id, 'company_ids': (self.company | self.company_2).ids})
        with self.assertRaisesRegex(UserError, "you don't have access to that company"):
            subscription = self.subscription.with_context(allowed_company_ids=[self.company_2.id])
            subscription._create_invoices()

        # enable "company_1" in context, we should be able to invoice now
        subscription = self.subscription.with_context(allowed_company_ids=(self.company | self.company_2).ids)
        subscription._create_invoices()

    def test_manual_switching_company_of_subscription(self):
        """Test moving subscription to a different company (ex. "Self Odoo" from Odoo SA to subsidiary.)"""

        # drop administration privileges from accountman
        self.env.user.groups_id -= (
            self.env.ref("base.group_erp_manager") | self.env.ref("base.group_system")
        )
        self.assertEqual(self.subscription.company_id, self.company)
        self.assertEqual(self.subscription.order_line.tax_id, self.tax_10)  # tax 10% of company_1
        self.assertEqual(self.subscription.warehouse_id.company_id, self.company)
        self.assertFalse(self.env.su)
        self.subscription.action_confirm()

        # Restrict accountant access to only "company_2", the user should be able to switch
        # the subscription to it and taxes should be correctly recomputed.
        self.env.user.write({'company_id': self.company_2.id, 'company_ids': self.company_2.ids})
        self.subscription.write({
            'company_id': self.company_2.id,
            'order_line': [
                (1, line.id, {'tax_id': [(6, 0, [self.company_2_tax_10.id])]})
                for line in self.subscription.order_line
            ]
        })

        # flush and clear cache, ensure we only have taxes from company_2
        self.env.flush_all()
        self.env.clear()
        self.assertEqual(
            self.subscription.order_line.sudo().tax_id, self.company_2_tax_10,
            "There should only be taxes from company_2 on the subscription"
        )
        self.assertEqual(
            self.subscription.warehouse_id.sudo().company_id, self.company_2,
            "The warehouse of the subscription should be the one of company_2"
        )

    def test_manual_invoicing_subscription_with_payment_token(self):
        """When subscription has a payment token's from a different company, invoicing should be created in that company."""

        # Change partner country to test fiscal position mapping
        self.subscription.partner_id.country_id = self.env.ref("base.ke")

        # Create a payment token for the 2nd company and update & confirm the subscription
        payment_token = self._create_token(provider_id=self.acquirer_2.id)
        self.subscription.write({
            'payment_token_id': payment_token.id,
        })
        self.subscription.action_confirm()

        # Create an invoice, that invoice should be on the company of the payment token
        # (so here "company_2" and NOT "company_1")
        self.subscription._create_invoices()
        self.assertEqual(len(self.subscription.invoice_ids), 1, "We should have create an invoice")
        invoice = self.subscription.invoice_ids
        self.assertEqual(invoice.state, "draft", "Manually created invoice should be in Draft state")
        self.assertEqual(invoice.company_id, self.company_2, "The invoice's company should match the one of the payment token")
        self.assertEqual(invoice.fiscal_position_id, self.company_2_extracomm_fpos, "The fiscal position should be extra-comm.")
        self.assertEqual(invoice.amount_tax, 0, "Extra-comm, there should be not tax.")
        self.assertTrue(all(l.tax_ids == self.company_2_tax_0 for l in invoice.invoice_line_ids))
        self.assertEqual(invoice.amount_untaxed, self.subscription.amount_untaxed, "Same untaxed amount as the subscription")

    def test_manual_invoicing_subscription_with_payment_token_dmcc(self):
        """When subscription has a payment token's from a different company, for DMCC we should always created invoice in DMCC."""

        # Force Odoo DMCC xmlid reference to current company
        dmcc_imd = self.env["ir.model.data"].search([("module", "=", "openerp_enterprise"), ("name", "=", "company_dmcc")])
        dmcc_imd.res_id = self.company.id
        dmcc_imd.flush_recordset()
        # clear caches and force cache clearing after test so that we reset to initial value
        self.env.registry.clear_cache()
        self.addCleanup(self.env.registry.clear_cache)

        # Change partner country to test fiscal position mapping
        self.subscription.partner_id.country_id = self.env.ref("base.ke")

        # Create a payment token for the 2nd company and update & confirm the subscription
        payment_token = self._create_token(provider_id=self.acquirer_2.id)
        self.subscription.write({
            'payment_token_id': payment_token.id,
        })
        self.subscription.action_confirm()

        # Create an invoice, that invoice should be on the original company DMMC whatever the token
        # says (so here "company_1" in all cases)
        self.subscription._create_invoices()
        self.assertEqual(len(self.subscription.invoice_ids), 1, "We should have create an invoice")
        invoice = self.subscription.invoice_ids
        self.assertEqual(invoice.state, "draft", "Manually created invoice should be in Draft state")
        self.assertEqual(invoice.company_id, self.company, "The invoice's company should still be DMCC")
        self.assertEqual(invoice.fiscal_position_id, self.subscription.fiscal_position_id, "The fiscal position should be same are order")
        self.assertNotEqual(invoice.amount_tax, 0, "There should be some taxes.")
        self.assertTrue(all(l.tax_ids == self.tax_10 for l in invoice.invoice_line_ids))
        self.assertEqual(invoice.amount_untaxed, self.subscription.amount_untaxed, "Same untaxed amount as the subscription")

    def test_auto_invoicing_subscription_with_payment_token(self):
        """When subscription has a payment token's from a different company, invoicing should be created in that company."""

        # Change partner country to test fiscal position mapping
        self.subscription.partner_id.country_id = self.env.ref("base.ke")

        # Create a payment token for the 2nd company and update & confirm the subscription
        payment_token = self._create_token(provider_id=self.acquirer_2.id)
        self.subscription.write({
            'payment_token_id': payment_token.id,
        })

        # Run the cron to create an invoice, that invoice should be on the company of the payment
        # token(so here "company_2" and NOT "company_1")
        with patch('odoo.addons.sale_subscription.models.sale_order.SaleOrder._do_payment',
                   wraps=self._mock_subscription_do_payment_and_reconcile),\
            patch('odoo.addons.sale_subscription.models.sale_order.SaleOrder._send_success_mail',
                  wraps=self._mock_subscription_send_success_mail):

            self.subscription.action_confirm()
            self.mock_send_success_count = 0
            self.env['sale.order'].sudo()._cron_recurring_create_invoice()

        tx = self.subscription.transaction_ids[-1:]
        self.assertEqual(tx.company_id, self.company_2, "Payment transaction should be on company 2")
        self.assertEqual(len(self.subscription.invoice_ids), 1, "We should have create an invoice")
        invoice = self.subscription.invoice_ids
        self.assertEqual(invoice.state, "posted", "Invoice should be posted")
        self.assertEqual(invoice.company_id, self.company_2, "The invoice's company should match the one of the payment token")
        self.assertEqual(invoice.fiscal_position_id, self.company_2_extracomm_fpos, "The fiscal position should be extra-comm.")
        self.assertEqual(invoice.amount_tax, 0, "Extra-comm, there should be not tax.")
        self.assertTrue(all(l.tax_ids == self.company_2_tax_0 for l in invoice.invoice_line_ids))
        self.assertEqual(invoice.amount_untaxed, self.subscription.amount_untaxed, "Same untaxed amount as the subscription")
        self.assertEqual(invoice.payment_state, "not_paid", "Invoice should still be unpaid")
        self.assertEqual(
            self.subscription.next_invoice_date,
            self.subscription.start_date +  self.subscription.plan_id.billing_period,
            "Next invoice date should have been updated and be 1 month after start of subscription."
        )

        # Force processing of transaction (normally done by cron), tx payment and invoice should
        # be reconciled.
        tx._finalize_post_processing()
        self.assertEqual(invoice.payment_state, "in_payment", "Invoice should now be 'In Payment'")
        self.assertEqual(invoice.amount_residual, 0.0, "Invoice is fully paid.")

    def test_auto_invoicing_subscription_with_payment_token_dmcc(self):
        """When subscription has a payment token's from a different company, invoicing should be created in that company."""

        # Force Odoo DMCC xmlid reference to current company
        dmcc_imd = self.env["ir.model.data"].search([("module", "=", "openerp_enterprise"), ("name", "=", "company_dmcc")])
        dmcc_imd.res_id = self.company.id
        dmcc_imd.flush_recordset()
        # clear caches and force cache clearing after test so that we reset to initial value
        self.env.registry.clear_cache()
        self.addCleanup(self.env.registry.clear_cache)

        # Change partner country to test fiscal position mapping
        self.subscription.partner_id.country_id = self.env.ref("base.ke")

        # Create a payment token for the 2nd company and update & confirm the subscription
        payment_token = self._create_token(provider_id=self.acquirer_2.id)
        self.subscription.write({
            'payment_token_id': payment_token.id,
        })

        # Run the cron to create an invoice, that invoice should be on the company of the payment
        # token(so here "company_2" and NOT "company_1")
        with patch('odoo.addons.sale_subscription.models.sale_order.SaleOrder._do_payment',
                   wraps=self._mock_subscription_do_payment_and_reconcile),\
            patch('odoo.addons.sale_subscription.models.sale_order.SaleOrder._send_success_mail',
                  wraps=self._mock_subscription_send_success_mail):

            self.subscription.action_confirm()
            self.mock_send_success_count = 0
            self.env['sale.order'].sudo()._cron_recurring_create_invoice()

        tx = self.subscription.transaction_ids[-1:]
        self.assertEqual(tx.company_id, self.company_2, "Payment transaction should be on company 2")
        self.assertEqual(len(self.subscription.invoice_ids), 1, "We should have create an invoice")
        invoice = self.subscription.invoice_ids
        self.assertEqual(invoice.state, "posted", "Invoice should be posted")
        self.assertEqual(invoice.company_id, self.company, "The invoice's company should still be DMCC")
        self.assertEqual(invoice.fiscal_position_id, self.subscription.fiscal_position_id, "The fiscal position should be same are order")
        self.assertNotEqual(invoice.amount_tax, 0, "There should be some taxes.")
        self.assertTrue(all(l.tax_ids == self.tax_10 for l in invoice.invoice_line_ids))
        self.assertEqual(invoice.amount_untaxed, self.subscription.amount_untaxed, "Same untaxed amount as the subscription")
        self.assertEqual(invoice.payment_state, "not_paid", "Invoice should still be unpaid")
        self.assertEqual(
            self.subscription.next_invoice_date,
            self.subscription.start_date + self.subscription.plan_id.billing_period,
            "Next invoice date should have been updated and be 1 month after start of subscription."
        )

        # Force processing of transaction (normally done by cron). TX's payment and invoice should
        # NOT be reconciled as they are on a differ company.
        tx._finalize_post_processing()
        self.assertEqual(invoice.payment_state, "not_paid", "Invoice should now be 'In Payment'")
        self.assertEqual(invoice.amount_residual, invoice.amount_total, "Invoice is not be paid.")

    def test_payment_token_auto_archive_for_expired_card(self):
        self.subscription.write({'payment_token_id': self.payment_token.id})
        self.subscription.action_confirm()

        def _mock_subscription_do_payment_rejected(payment_method, invoice, auto_commit=False):
            tx = self._mock_subscription_do_payment(payment_method, invoice, auto_commit=auto_commit)
            tx.state = "pending"
            tx._set_error("Payment declined")
            tx.ogone_error_code = 30331001  # card-expired
            tx.env.cr.flush()  # simulate commit after sucessfull `_do_payment()`
            return tx

        with patch('odoo.addons.sale_subscription.models.sale_order.SaleOrder._do_payment',
                   wraps=_mock_subscription_do_payment_rejected),\
             self.mock_mail_gateway():
            self.subscription._create_recurring_invoice()

        self.assertEqual(self.payment_token.active, False, "The payment token should have been archived")
        self.assertFalse(self.subscription.payment_token_id, "The payment token should have been removed from the subscription")


    def test_auto_payment_with_token(self):
        # In standard, this method test cron (automatic) invoicing/payment for subscription,
        # both for sucessfull payments [case 1] and rejected payments [case 2].
        #   Specifically [case 2] tests for possible cron infinite loops when number of
        # subscriptions to invoice exceeds the batch size and thus use ir.cron trigger()
        # (test initially introduced in odoo/odoo@bb6176bbe5b7, `invalid_payment_tag` being
        # now `is_batch` field), but as we're only calling sale_subscription's
        # `_create_recurring_invoice()` method with non-empty recordset, we're never going to
        # trigger() (`need_cron_trigger` is always False)
        #   As for [case 1], we have other tests already covering this case, so we skip this test.
        raise self.skipTest("skip test for openerp_enterprise")

    def test_close_unpaid_contracts(self):

        quotation_template_yearly = self.env.ref('openerp_enterprise.custom_template_odoo_yearly')
        quotation_template_yearly.plan_id = self.plan_year

        with self._mock_enable_order_payment_with_token(), \
                self._patch_model_default('sale.order', 'sale_order_template_id', quotation_template_yearly):
            super().test_close_unpaid_contracts()

    def test_close_unpaid_contracts_bis(self):
        # For the purpose of the test, force "non-automatic" cron behavior
        # to allow creating invoice for yearly subs. wo/ token
        @monkey_patch(self.env.registry["sale.order"], "_filter_subscription_to_invoice")
        def _patched_filter_subscription_to_invoice(self, *args, **kwargs):
            return _patched_filter_subscription_to_invoice.super(self, automatic=False)
        self.addCleanup(_patched_filter_subscription_to_invoice.reset)

        super().test_close_unpaid_contracts_bis()

    def test_subscription_compatible_acquirers(self):
        """Test that acquirers are correctly filters according to their recurrence."""
        def _get_compatible_acquirers(subscription=None):
            subscription = subscription or self.subscription
            return self.env['payment.provider'].sudo()._get_compatible_providers(
                subscription.company_id.id,
                subscription.partner_id.id,
                subscription.amount_total,
                currency_id=subscription.currency_id.id,
                sale_order_id=subscription.id,
            )

        def _prepare_acquirer(allowed_freqs, **extra_values):
            new_acquirer = self.provider.copy(dict({
                f"allow_{freq}_payments": bool(freq in allowed_freqs)
                for freq in ['one_shot', 'monthly', 'yearly']
            }, **extra_values))
            new_acquirer.state = "test"
            return new_acquirer

        # Create provider with tokenization support
        acquirer_all_freq = _prepare_acquirer(['one_shot', 'monthly', 'yearly'], allow_tokenization=True)
        acquirer_monthly_only_freq = _prepare_acquirer(['monthly'], allow_tokenization=True)
        acquirer_yearly_only_freq = _prepare_acquirer(['yearly'], allow_tokenization=True)
        acquirer_one_shot_only_freq = _prepare_acquirer(['one_shot'], allow_tokenization=True)

        # Create providers without tokenization support
        acquirer_monthly_only_freq_notok = _prepare_acquirer(['monthly'], allow_tokenization=False)
        acquirer_yearly_only_freq_notok = _prepare_acquirer(['yearly'], allow_tokenization=False)

        all_acquirers = (
            acquirer_all_freq |
            acquirer_monthly_only_freq | acquirer_monthly_only_freq_notok |
            acquirer_yearly_only_freq | acquirer_yearly_only_freq_notok |
            acquirer_one_shot_only_freq
        )

        # check that we correctly filter payment provider depending on the recurrence
        self.subscription.sale_order_template_id.plan_id = self.env.ref("openerp_enterprise.subscription_plan_odoo_monthly_plan")
        self.subscription.plan_id = self.env.ref("openerp_enterprise.subscription_plan_odoo_monthly_plan")
        self.assertEqual(
            _get_compatible_acquirers() & all_acquirers,
            acquirer_all_freq | acquirer_monthly_only_freq,
            "We should only get provider with monthly allowed payments and supporting tokenization."
        )

        self.subscription.sale_order_template_id.plan_id = self.env.ref("openerp_enterprise.subscription_plan_odoo_yearly_plan")
        self.subscription.plan_id = self.env.ref("openerp_enterprise.subscription_plan_odoo_yearly_plan")
        self.assertEqual(
            _get_compatible_acquirers() & all_acquirers,
            acquirer_all_freq | acquirer_yearly_only_freq | acquirer_yearly_only_freq_notok,
            "We should only get provider with yearly allowed payments (with or without tokenization).",
        )

        # for upsell or normal sale order, we shouldn't limit based on the recurrence
        self.subscription.action_confirm()
        self.subscription._create_invoices()._post()
        action = self.subscription.prepare_upsell_order()
        upsell_order = self.env["sale.order"].browse(action["res_id"])
        self.assertEqual(
            _get_compatible_acquirers(subscription=upsell_order) & all_acquirers,
            all_acquirers,
            "We should not limit provider based on allowed payments frequency."
        )

    def test_subscription_manage_payment_methods(self):
        """Ensure that subscription 'Manage Payments methods' shows those of the sub's company."""
        self.subscription.company_id = self.company_2
        self.subscription.action_confirm()
        self.assertEqual(self.user_portal.company_id, self.company)

        # Allow both company providers to be used to pay for subscriptions
        (self.provider | self.acquirer_2).write({
            'allow_one_shot_payments': True,
            'allow_monthly_payments': True,
            'allow_yearly_payments': True,
            'state': 'test',
        })

        # User is on company_1, subscription on company_2: ensure we show
        # users the providers of the subscription (i.e company_2)
        providers = self.env["payment.provider"].sudo()._get_compatible_providers(
            self.user_portal.company_id.id,
            self.user_portal.partner_id.id,
            0,
            is_validation=True,
            manage_subscription=True,
            sale_order_id=self.subscription.id,
        )
        self.assertEqual(
            providers, self.acquirer_2,
            "When managing subscription payment methods, we should always return"
            "payment providers of the subscription's company",
        )


class SubscriptionCommon(EnterpriseDatabaseCommon):

    @classmethod
    def init_master_data(cls):
        group_user = cls.env.ref('base.group_user')
        see_subscription = cls.env.ref('sales_team.group_sale_salesman')
        see_database = cls.env.ref('openerp_enterprise.group_all_readonly')

        cls.see_sub_user = cls.env['res.users'].create({
            'login': 'ro_subscription',
            'name': 'RO Subscription',
            'groups_id': [(6, 0, [group_user.id, see_subscription.id])]
        })
        cls.no_see_sub_user = cls.env['res.users'].create({
            'login': 'no_subscription',
            'name': 'No Subscription',
            'groups_id': [(6, 0, [group_user.id])]
        })
        cls.salesman_user = cls.env['res.users'].create({
            'login': 'salesman_subscription',
            'name': 'Salesman Subscription',
            'email': 'salesman-user@example.net',
            'groups_id': [(6, 0, [group_user.id, see_subscription.id, see_database.id])]
        })
        cls.customer = cls.env['res.partner'].create({
            'name': 'Subscription Customer',
            'email': 'enterprise-customer@example.net',
            'country_id': cls.env.ref('base.be').id,
            'property_product_pricelist': cls.env.ref('openerp_enterprise.pl22_eu_high').id,
        })
        cls.payment_method = cls.env.ref('payment.payment_method_paypal')
        cls.token = cls.env['payment.token'].create({
            'payment_details': "Dummy Token",
            'partner_id': cls.customer.id,
            'payment_method_id':cls.payment_method.id,
            'provider_id': cls.env.ref('payment.payment_provider_paypal').id,
            'provider_ref': 'TOPSECRET',
        })
        cls.success_pack_product = cls.env['product.product'].create({
            'name': 'Huge Success Pack (like very huge)',
            'categ_id': cls.env.ref('openerp_enterprise.success_packs_categ_id').id,
            'type': 'service',
            'purchase_ok': True,
            'sale_ok': True,
            'service_policy': 'ordered_prepaid',
            'recurring_invoice': False,
        })
        cls.certification_product = cls.env["product.product"].create({
            'name': "Certification Odoo Ultimate Inter-Languages Localizations Extended",
            'categ_id': cls.env.ref('openerp_enterprise.product_category_training').id,
            'detailed_type': 'course',
            'is_published': True,
            'list_price': 245.0,
            'sale_ok': True,
        })

    qt_standard_monthly = 'openerp_enterprise.standard_template_odoo_monthly'
    qt_standard_yearly = 'openerp_enterprise.standard_template_odoo_yearly'
    qt_custom_monthly = 'openerp_enterprise.custom_template_odoo_monthly'
    qt_custom_yearly = 'openerp_enterprise.custom_template_odoo_yearly'
    qt_partnership = 'openerp_enterprise.quotation_template_odoo_partnership'
    qt_partnership_learning = 'openerp_enterprise.quotation_template_odoo_learning_partnership'

    def _create_so_from_template(self, partner, quotation_template_xmlid=False, quotation_template=False):
        so_tmp_id = self.env.ref(quotation_template_xmlid).id if quotation_template_xmlid else quotation_template.id
        order = self.env['sale.order'].create({
            'partner_id': partner.id,
            'sale_order_template_id': so_tmp_id,
        })
        order._onchange_sale_order_template_id()
        return order

    def init_so(self):
        self.so = self._create_so_from_template(self.customer, self.quotation_template)

    def init_data(self):
        self.database_create_date = fields.Datetime.to_string(now())
        self.init_so()
        self.so.action_confirm()
        subscriptions = self.so
        odoo_user_standard = self.env.ref('openerp_enterprise.product_odoo_user_standard')
        odoo_user_custom = self.env.ref('openerp_enterprise.product_odoo_user_custom')
        odoo_users = odoo_user_standard | odoo_user_custom
        self.main_subscription = subscriptions.filtered(lambda x: odoo_users & x.order_line.mapped('product_id'))
        self.code = self.main_subscription.client_order_ref

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.init_master_data()

    def setUp(self):
        super().setUp()
        # Set a fake datetime, if a ping occur during the same second
        # The second ping will have a ping time with at least one more second
        # That ensure that the last ping will be the last_ping_id of the database
#        self.env['openerp.enterprise.ping']._fields['ping_time'].default = lambda x: FakeDatetime.now()
        self.init_data()

    def tearDown(self):
        super().tearDown()
        self.env['openerp.enterprise.ping']._fields['ping_time'].default = lambda x: datetime.datetime.now()

    def send_ping(self, db_uuid, modules, db_create_date=None, errors=False, code=False, payload_update=None, fromip=None):
        # Note: force database creation date to test's start date
        return super().send_ping(db_uuid, modules, db_create_date=self.database_create_date,
                                 errors=errors, code=code, payload_update=payload_update, fromip=fromip)

    def upsell(self, contract):
        action = contract.prepare_upsell_order()
        return self.env['sale.order'].browse(action['res_id'])

    def renew(self, contract):
        action = contract.prepare_renewal_order()
        return self.env["sale.order"].browse(action['res_id'])


class SubscriptionYearly(SubscriptionCommon):
    quotation_template = SubscriptionCommon.qt_standard_yearly

    def test_users_upsell(self):
        """ Test upsell of user and cancel the upsell """
        self.main_subscription._create_invoices()._post()

        db_uuid = generate_uuid()
        self.send_ping(db_uuid, {})
        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        database.subscription_id = self.main_subscription
        # Pïng should update draft and close Yearly Subscription
        res = self.send_ping(db_uuid, {}, code=self.code, payload_update={
            'nbr_users': 10,
            'nbr_active_users': 10,
        })
        self.assertEqual(res['enterprise_info']['expiration_reason'], 'upsell', "Need to upsell Users")
        self.assertEqual(self.main_subscription.to_renew, True, "Status should be pending")
        self.assertTrue(self.main_subscription.upsell_date, "Upsell date should be set")
        # Trigger Upsell sale order
        upsell_so = self.upsell(self.main_subscription)
        self.assertEqual(self.main_subscription.upsell_order_id, upsell_so, "Upsell order should be properly set")
        # Remove Users that generate the upsell
        res = self.send_ping(db_uuid, {}, code=self.code, payload_update={
            'nbr_users': 0,
            'nbr_active_users': 0,
        })
        self.assertEqual(res['enterprise_info']['expiration_reason'], 'renewal', "No need to upsell anymore")
        self.assertEqual(self.main_subscription.subscription_state, '3_progress', "Status should be open")
        self.assertFalse(self.main_subscription.upsell_order_id, "Upsell order should be cancelled")
        self.assertFalse(self.main_subscription.upsell_date, "Upsell date should be removed")

    def test_iotbox_upsell(self):
        """ Test upsell of subscription when new IoT box is registered. """
        iot_box_product = self.env.ref('openerp_enterprise.product_iot_box')

        self.assertNotIn(iot_box_product, self.main_subscription.order_line.product_id)
        self.main_subscription._create_invoices()._post()

        db_uuid = generate_uuid()
        res = self.send_ping(db_uuid, {}, payload_update={
            'IoTBox': 1,
            'enterprise_code': self.main_subscription.client_order_ref,
        })
        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        self.assertEqual(database.iot_box_count, 1)
        self.assertEqual(
            res['enterprise_info']['expiration_reason'], 'upsell',
            "We need to upsell IoT Boxes"
        )
        self.assertEqual(self.main_subscription.to_renew, True, "Status should be pending")
        self.assertTrue(self.main_subscription.upsell_date, "Upsell date should be set")

        # Trigger a manual upsell (should contain new "IoT Box" sub. service line)
        upsell_so = self.upsell(self.main_subscription)
        self.assertIn(iot_box_product, upsell_so.order_line.product_id)
        upsell_so.action_confirm()

        # Check are IoT Box are now added to subscription
        iot_box_line = self.main_subscription.order_line.filtered(lambda x: x.product_id == iot_box_product)
        self.assertEqual(len(iot_box_line), 1, "There should be a single Iot Box line")
        self.assertEqual(iot_box_line.product_id, iot_box_line.product_id)
        self.assertEqual(iot_box_line.product_uom_qty, 1)

    def test_sub_multiyear_invoice_user_discount_deferred_dates(self):
        """Ensure that description and deferred dates of First Year discount invoice lines are correct."""
        self.customer.lang = 'en_US'
        std_user_discount_prod = self.env.ref("openerp_enterprise.product_odoo_user_standard_discount")
        cust_user_discount_prod = self.env.ref("openerp_enterprise.product_odoo_user_custom_discount")

        with freeze_time('2024-01-10'):
            sub = self._create_so_from_template(self.customer, 'openerp_enterprise.standard_template_odoo_2_years')
            sub.action_confirm()
            invoice = sub._create_invoices()._post()
            self.assertEqual(len(invoice), 1)
        discount_line = invoice.invoice_line_ids.filtered(
            lambda l: l.product_id == std_user_discount_prod
        )
        self.assertEqual(discount_line.deferred_start_date, datetime.date(2024, 1, 10))
        self.assertEqual(discount_line.deferred_end_date, datetime.date(2025, 1, 9))
        description = discount_line.name
        self.assertTrue(
            description.endswith('1 Years\n01/10/2024 to 01/09/2025'),
            "First Year discount invoice lines should indicate correct period duration and dates",
        )

        # 4 months later we upsell the subscription with a change on plan
        # (no database linked, fallback to onpremise so switch: standard -> custom)
        with freeze_time('2024-05-15'):
            upsell = self.upsell(sub)
            upsell.action_confirm()
            upsell_invoice = upsell._create_invoices(final=True)
            self.assertEqual(len(upsell_invoice), 1)
        std_upsell_discount_line = upsell_invoice.invoice_line_ids.filtered(
            lambda l: l.product_id == std_user_discount_prod
        )
        self.assertEqual(std_upsell_discount_line.deferred_start_date, datetime.date(2024, 5, 15))
        self.assertEqual(std_upsell_discount_line.deferred_end_date, datetime.date(2025, 1, 9))
        self.assertTrue(
            std_upsell_discount_line.name.endswith('1 Years\n05/15/2024 to 01/09/2025'),
            "First Year discount invoice lines should indicate correct period duration and dates",
        )
        cust_upsell_discount_line = upsell_invoice.invoice_line_ids.filtered(
            lambda l: l.product_id == cust_user_discount_prod
        )
        self.assertEqual(cust_upsell_discount_line.deferred_start_date, datetime.date(2024, 5, 15))
        self.assertEqual(cust_upsell_discount_line.deferred_end_date, datetime.date(2025, 1, 9))
        self.assertTrue(
            cust_upsell_discount_line.name.endswith('1 Years\n05/15/2024 to 01/09/2025'),
            "First Year discount invoice lines should indicate correct period duration and dates",
        )


class SubscriptionMergeDiscount(SubscriptionYearly):
    quotation_template = SubscriptionYearly.qt_standard_yearly

    def get_discount_products(self):
        return (
            self.env.ref('openerp_enterprise.product_official_partner_discount') |
            self.env.ref('openerp_enterprise.product_learning_partner_discount') |
            self.env.ref('openerp_enterprise.product_odoo_subscription_extra_discount') |
            self.env.ref('openerp_enterprise.product_multiyear_discount')
        )

    def get_discount_lines(self, price, qty=1):
        return [Command.create({
            'product_id': discount_product.id,
            'product_uom_qty': qty,
            'price_unit': price,
            'product_uom': self.main_subscription.sale_order_template_id.get_product_uom(discount_product).id,
        }) for discount_product in self.get_discount_products()]

    def upsell_discount(self, price, qty=1):
        upsell = self.upsell(self.so)
        upsell.order_line.unlink()
        upsell.write({'order_line': self.get_discount_lines(price, qty=qty)})
        upsell.action_confirm()

    def test_merge_discount_price(self):
        self.main_subscription.write({'order_line': self.get_discount_lines(-2)})
        self.main_subscription._create_invoices()._post()
        total = self.main_subscription.amount_untaxed
        amount = -1
        qty = 1
        self.upsell_discount(amount, qty=qty)

        for product_discount in self.get_discount_products():
            discount_line = self.main_subscription.order_line.filtered(lambda l, p=product_discount: l.product_id == p)
            self.assertEqual(len(discount_line), 1)
            self.assertEqual(discount_line.product_uom_qty, 1)
            self.assertEqual(discount_line.price_unit, -3)
        self.assertEqual(self.main_subscription.amount_untaxed, total + amount * qty * 4)

    def test_merge_discount_qty(self):
        self.main_subscription.write({'order_line': self.get_discount_lines(-2, qty=2)})
        self.main_subscription._create_invoices()._post()
        total = self.main_subscription.amount_untaxed
        amount = -2
        qty = 4
        self.upsell_discount(amount, qty=qty)

        for product_discount in self.get_discount_products():
            discount_line = self.main_subscription.order_line.filtered(lambda l, p=product_discount: l.product_id == p)
            self.assertEqual(len(discount_line), 1)
            self.assertEqual(discount_line.product_uom_qty, 6)
            self.assertEqual(discount_line.price_unit, -2)
        self.assertEqual(self.main_subscription.amount_untaxed, total + amount * qty * 4)

    def test_no_merge_discount(self):
        self.main_subscription.write({'order_line': self.get_discount_lines(-2, qty=1)})
        self.main_subscription._create_invoices()._post()
        total = self.main_subscription.amount_untaxed
        amount = -1
        qty = 2
        self.upsell_discount(amount, qty=2)

        for product_discount in self.get_discount_products():
            discount_line = self.main_subscription.order_line.filtered(lambda l, p=product_discount: l.product_id == p)
            self.assertEqual(len(discount_line), 2)

        self.assertEqual(self.main_subscription.amount_untaxed, total + amount * qty * 4)


class TestSubscriptionMonthly(SubscriptionCommon):
    quotation_template = SubscriptionCommon.qt_standard_monthly

    def init_data(self):
        super().init_data()
        self.main_subscription = self.so
        self.code = self.main_subscription.client_order_ref

    def _create_so_from_template(self, partner, quotation_template):
        order = self.env['sale.order'].create({
            'partner_id': partner.id,
            'sale_order_template_id': self.env.ref(quotation_template).id,
        })
        order._onchange_sale_order_template_id()
        return order

    def test_database_link_to_subscription_update_users_count(self):
        """ Test linking a database to a monthly subscription (manual computation) update the user count correctly"""

        odoo_user_prod_categ = self.env.ref("openerp_enterprise.product_category_odoo_users")
        odoo_user_prod = self.env.ref("openerp_enterprise.product_odoo_user_standard")
        odoo_user_ini_discount_prod = self.env.ref("openerp_enterprise.product_odoo_user_standard_discount")

        # we sold a subscription for 1 user, database had more active users in the meantime,
        # check that we increase the users count
        subscription = self._create_so_from_template(self.customer, self.quotation_template)
        lines_odoo_user = subscription.order_line.filtered(lambda x: x.product_id.categ_id == odoo_user_prod_categ)
        lines_odoo_user.write({
            'product_uom_qty': 1.0,
            'computation': 'manual',
        })
        subscription.action_confirm()
        db_uuid = generate_uuid()
        self.send_ping(db_uuid, {}, payload_update=dict(nbr_users=2, nbr_active_users=2))
        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        database.subscription_id = subscription
        self.assertRecordValues(
            lines_odoo_user,
            [
                {'product_id': odoo_user_prod.id, 'product_uom_qty': 2.0},
                {'product_id': odoo_user_ini_discount_prod.id, 'product_uom_qty': 1.0},  # no update
            ]
        )

        # we sold a subscription for 10 users, database still have less active users, check that
        # we do not decrease the user count.
        subscription2 = self._create_so_from_template(self.customer, self.quotation_template)
        lines_odoo_user2 = subscription2.order_line.filtered(lambda x: x.product_id.categ_id == odoo_user_prod_categ)
        lines_odoo_user2.product_uom_qty = 10.0
        subscription2.action_confirm()
        db_uuid2 = generate_uuid()
        self.send_ping(db_uuid2, {}, payload_update=dict(nbr_users=4, nbr_active_users=4))
        database2 = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid2)])
        database2.subscription_id = subscription2
        self.assertRecordValues(
            lines_odoo_user2,
            [
                {'product_id': odoo_user_prod.id, 'product_uom_qty': 10.0},
                {'product_id': odoo_user_ini_discount_prod.id, 'product_uom_qty': 10.0},
            ]
        )

    def test_iotbox_upsell(self):
        """ Test auto. upsell of subscription when new IoT box is registered. """
        iot_box_product = self.env.ref('openerp_enterprise.product_iot_box')

        self.assertNotIn(iot_box_product, self.main_subscription.order_line.product_id)
        self.main_subscription._create_invoices()._post()

        db_uuid = generate_uuid()
        res = self.send_ping(db_uuid, {}, payload_update={
            'IoTBox': 1,
            'enterprise_code': self.main_subscription.client_order_ref,
        })
        self.assertEqual(res['enterprise_info']['expiration_reason'], 'renewal')
        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        self.assertEqual(database.iot_box_count, 1)

        # Check are IoT Box sub. service have been (automatically) added to subscription
        iot_box_line = self.main_subscription.order_line.filtered(lambda x: x.product_id == iot_box_product)
        self.assertEqual(
            len(iot_box_line), 1,
            "A IoT Box line should have been automatically added by the ping."
        )
        self.assertEqual(iot_box_line.product_id, iot_box_line.product_id)
        self.assertEqual(iot_box_line.product_uom_qty, 1)

    def test_users_plan_mismatch(self):
        """Test subscription plan auto-switch"""
        odoo_user_standard = self.env.ref('openerp_enterprise.product_odoo_user_standard')
        odoo_user_custom = self.env.ref('openerp_enterprise.product_odoo_user_custom')
        standard_yearly_plan_template = self.env.ref('openerp_enterprise.standard_template_odoo_yearly')

        # Create "Odoo Monthly Standard" subscription
        sub = self.env['sale.order'].with_user(self.salesman_user).create({
            'partner_id': self.customer.id,
            'sale_order_template_id': standard_yearly_plan_template.id,
        })
        sub._onchange_sale_order_template_id()

        # Try to change only 'Standard Plan' by 'Custom Plan'
        with self.assertRaises(UserError, msg="The product.*Custom Plan.*cannot be added to a standard plan"):
            sub_users_line = sub.order_line.filtered(lambda l: l.product_id & odoo_user_standard)
            sub.write({
                'order_line': [
                    (1, sub_users_line.id, {'product_id': odoo_user_custom.id}),
                ]
            })

    def test_subscription_name_search(self):
        """Ensure that we can find subscription using it's ent. code, order name or both."""
        sub = self.main_subscription
        self.assertTrue(sub.client_order_ref)
        self.assertTrue(sub.name)
        self.assertEqual(sub.display_name, "%s - %s" % (sub.client_order_ref, sub.name))
        name_search = self.env["sale.order"].name_search
        self.assertIn((sub.id, sub.display_name), name_search(sub.client_order_ref))
        self.assertIn((sub.id, sub.display_name), name_search(sub.name))
        self.assertIn((sub.id, sub.display_name), name_search(sub.display_name))

class CheckPingAndSpoof(HttpCase, SubscriptionCommon):

    quotation_template = SubscriptionCommon.qt_standard_monthly

    @classmethod
    def init_master_data(cls):
        super().init_master_data()
        cls.customer_user = cls.env['res.users'].create({
            'login': cls.customer.email,
            'password': cls.customer.email,
            'partner_id': cls.customer.id,
            'groups_id': [(6, 0, cls.env.ref('base.group_portal').ids)],
        })

    def init_data(self):
        super().init_data()
        self.main_subscription.is_staff_restricted = True

    def test_change_sub(self):
        """
            Check the ping is still working with restricted subscription
            1) Create a new DB with a restricted subscription
            2) Change to another restricted subscription
        """
        db_uuid = generate_uuid()
        self.send_ping(db_uuid, {})
        db = self.env["openerp.enterprise.database"].search([('db_uuid', '=', db_uuid)])
        new_sub = self.main_subscription.copy(default={"client_order_ref": 'NEW_SUB_01'})
        new_sub.action_confirm()
        # Pïng should update draft and close Yearly Subscription
        res = self.send_ping(db_uuid, {}, code=self.code, payload_update={
            'nbr_users': 10,
            'nbr_active_users': 10,
        })
        db.invalidate_recordset(['spoof_ids'])  # spoof create in separate tx, force reload from db
        self.assertEqual(len(db.spoof_ids), 1)
        last_spoof = db.spoof_ids[:1]
        self.assertEqual(last_spoof.action, 'subscription')
        self.assertEqual(
            last_spoof.subscription_id, self.main_subscription,
            "Setting a database subscription should be logged with the new subscription",
        )
        self.assertIn(self.main_subscription.client_order_ref, last_spoof.reason)

        database = self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        res = self.send_ping(db_uuid, {}, code=new_sub.client_order_ref, payload_update={
            'nbr_users': 10,
            'nbr_active_users': 10,
        })
        self.assertEqual(database.subscription_id, new_sub)
        db.invalidate_recordset(['spoof_ids'])  # spoof create in separate tx, force reload from db
        self.assertEqual(len(db.spoof_ids), 2)
        last_spoof = db.spoof_ids[:1]
        self.assertEqual(last_spoof.action, 'subscription')
        self.assertEqual(
            last_spoof.subscription_id, new_sub,
            "Setting a database subscription should be logged with the new subscription",
        )
        self.assertIn(new_sub.client_order_ref, last_spoof.reason)

        # unlink database from subscription
        db.subscription_id = False
        db.invalidate_recordset(['spoof_ids'])  # spoof create in separate tx, force reload from db
        self.assertEqual(len(db.spoof_ids), 3)
        last_spoof = db.spoof_ids[:1]
        self.assertEqual(last_spoof.action, 'subscription')
        self.assertFalse(
            last_spoof.subscription_id,
        )
        self.assertIn('sub: False', last_spoof.reason)

    def test_self_oe_buy(self):
        """Test self-oe contract creation"""
        db_uuid = generate_uuid()
        self.send_ping(db_uuid, {}, payload_update={
            'nbr_users': 2,
            'nbr_active_users': 2,
        })
        db = self.env["openerp.enterprise.database"].search([('db_uuid', '=', db_uuid)])
        self.assertFalse(db.subscription_id)
        self.assertFalse(db.spoof_ids)

        self.authenticate(self.customer.email, self.customer.email)

        # Customer try to buy a subscription for its database
        resp = self.url_open('/odoo-enterprise/upgrade-online/%s' % db_uuid, allow_redirects=False)
        resp.raise_for_status()
        self_sub1 = db.subscription_id
        self.assertTrue(self_sub1)
        self.assertEqual(self_sub1.state, 'draft')
        db.invalidate_recordset(['spoof_ids'])  # spoof create in separate tx, force reload from db
        self.assertEqual(len(db.spoof_ids), 1)
        last_spoof = db.spoof_ids[:1]
        self.assertEqual(last_spoof.action, "subscription")
        self.assertEqual(last_spoof.subscription_id, self_sub1)
        self.assertEqual(last_spoof.user_id, self.customer_user)
        self.assertIn("sub: %s" % self_sub1.name, last_spoof.reason)

        # Then try again
        resp = self.url_open('/odoo-enterprise/upgrade-online/%s' % db_uuid, allow_redirects=False)
        resp.raise_for_status()
        self_sub2 = db.subscription_id
        self.assertTrue(self_sub2)
        self.assertEqual(self_sub2.state, 'draft',
                         "There should be a new draft subscription assigned to the database")
        self.assertEqual(self_sub1.state, 'cancel',
                         "The preceding draft subscription should now be cancelled")
        db.invalidate_recordset(['spoof_ids'])  # spoof create in separate tx, force reload from db
        self.assertEqual(len(db.spoof_ids), 2)
        last_spoof = db.spoof_ids[:1]
        self.assertEqual(last_spoof.action, "subscription")
        self.assertEqual(last_spoof.subscription_id, self_sub2)
        self.assertEqual(last_spoof.user_id, self.customer_user)
        self.assertIn("sub: %s" % self_sub2.name, last_spoof.reason)


@tagged('post_install', '-at_install')
class TestOdooSubscription(SaasWorkerHttpCommon, SubscriptionCommon):
    quotation_template = SubscriptionCommon.qt_custom_yearly

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.integrator = cls.env["res.partner"].create({
            'name': 'My Odoo Integrator',
            'email': 'odoo-integrator@example.net',
        })
        cls.customer_user = new_test_user(
            cls.env, cls.customer.email,
            partner_id=cls.customer.id,
            groups='base.group_portal',
        )

    def _create_sh_subscription(self, template=None, workers=1, stagings=1, storage=1, dedicated=0, users=1, maintenance=False):
        order = self._create_so_from_template(self.customer, template or self.quotation_template)
        if not maintenance:
            order._get_maintenance_line().unlink()

        user_products = [
            self.env.ref('openerp_enterprise.product_odoo_user_custom'),
            self.env.ref('openerp_enterprise.product_odoo_user_custom_discount'),
        ]
        users_lines = order.order_line.filtered(lambda l: l.product_id in user_products)
        if not users:
            users_lines.unlink()
        else:
            users_lines.product_uom_qty = users

        # add Odoo.sh products
        with Form(order) as so_form:
            for (product, qty, so_field) in [
                (self.env.ref('openerp_enterprise.product_odoo_sh_extra_worker'), workers, 'odoo_sh_extra_workers'),
                (self.env.ref('openerp_enterprise.product_odoo_sh_extra_staging'), stagings, 'odoo_sh_extra_stagings'),
                (self.env.ref('openerp_enterprise.product_odoo_sh_extra_gb'), storage, 'odoo_sh_extra_gb'),
                (self.env.ref('openerp_enterprise.product_odoo_sh_dedicated'), dedicated, 'odoo_sh_extra_dedicated')
            ]:
                if qty:
                    with so_form.order_line.new() as new_line:
                        new_line.product_id = product
                        new_line.product_uom_qty = qty
                    setattr(so_form, so_field, qty)  # Form does not support so_form[so_field] = value
        return order

    def _sh_repo_ping_info(self, workers=1, stagings=1, storage=1, dedicated=0):
        return {
            'extra_workers': workers,
            'extra_stagings': stagings,
            'extra_GB': storage,
            'extra_dedicated': dedicated
        }

    @patch("odoo.addons.openerp_enterprise.models.openerp_enterprise.get_roles_from_ip")
    def test_paas_ping(self, rolesfromip_mock):
        rolesfromip_mock.return_value = {'paas_worker'}

        with freeze_time('2023-01-15'):
            # Set enterprise contract for 10 users
            subscription = self._create_sh_subscription(users=10, workers=2, stagings=3, storage=20)
            subscription.referrer_id = self.integrator
            subscription.action_confirm()
            subscription._create_invoices()
            subscription.invoice_ids._post()

            # Create database, ping (to link it to the subscription)
            database = self.create_database()
            self.send_ping(
                database.db_uuid, {'my_awesome_sh_custom': 4200},
                code=subscription.client_order_ref,
                payload_update={'nbr_users': 10, 'nbr_active_users': 5}
            )
            database.invalidate_recordset(['spoof_ids'])  # spoof create in separate tx, force reload from db
            self.assertEqual(len(database.spoof_ids), 1)
            last_spoof = database.spoof_ids[:1]
            self.assertEqual(last_spoof.action, "subscription")
            self.assertEqual(last_spoof.user_id.id, SUPERUSER_ID)
            self.assertEqual(
                last_spoof.subscription_id, subscription,
                "Setting a database subscription should be logged with the new subscription"
            )
            self.assertIn('sub: %s' % subscription.display_name, last_spoof.reason)

        self.assertFalse(subscription.need_odoo_sh_upsell())
        self.assertFalse(subscription.upsell_date)

        self.env.cr.flush()
        with self._mock_saas_config(from_role='paas_master'):
            resp = self.url_open(
                '/odoo-enterprise/paas-subscription-ping',
                auth=SaasAuth(),
                json={'params': {
                    'fetch_only': False,
                    'subscriptions': {
                        subscription.client_order_ref:
                            self._sh_repo_ping_info(workers=3, stagings=4, storage=38, dedicated=0),
                    }
                }}
            )
            result = resp.json().get('result')
            subscription.invalidate_recordset()  # ensure cache is clean
            # check that subscription is correctly categorized
            self.assertIn(subscription.client_order_ref, result['valid'])
            self.assertFalse(result['invalid'])
            self.assertFalse(result['trial'])

            self.assertEqual(result[subscription.client_order_ref], {
                'id': subscription.id,
                'enterprise_is_partnership': False,
                'customer_email': 'enterprise-customer@example.net',
                'partner_email': 'odoo-integrator@example.net',
                'users': 10.0,
                'periodicity': {'unit': 'year', 'duration': 1},
                'next_invoice_date': '2024-01-15',
                # values are the one on the sub before the paas ping
                'workers': 2.0,
                'staging': 3.0,
                'storage': 20.0,
                'dedicated': 0,
            })

            self.assertRecordValues(subscription, [{
                'odoo_sh_extra_workers': 3,
                'odoo_sh_extra_stagings': 4,
                'odoo_sh_extra_gb': 38,
                'odoo_sh_extra_dedicated': 0,
            }])
            self.assertTrue(subscription.need_odoo_sh_upsell(), "The subscription should be upselled")
            self.assertTrue(subscription.upsell_date, "An upsell date should have been set")

    @patch("odoo.addons.openerp_enterprise.models.openerp_enterprise.get_roles_from_ip")
    def test_paas_ping_separated_enterprise_sub_info(self, rolesfromip_mock):
        """Test paas ping in case enterprise sub is separated from Odoo.sh sub"""
        rolesfromip_mock.return_value = {'paas_worker'}

        yearly_template = 'openerp_enterprise.custom_template_odoo_yearly'
        monthly_template = 'openerp_enterprise.custom_template_odoo_monthly'

        # Create an Odoo.sh subscription (without Odoo users)
        with freeze_time("2023-01-15"):
            odoosh_subscription = self._create_sh_subscription(template=yearly_template, users=0, dedicated=1)
            odoosh_subscription.client_order_ref = 'M202303SHSUB001'
            odoosh_subscription.action_confirm()
            odoosh_subscription._create_invoices()
            odoosh_subscription.invoice_ids._post()

        # Create the enterprise subscription
        with freeze_time("2023-01-20"):
            ent_subscription = self._create_sh_subscription(
                template=monthly_template, users=10, workers=0, stagings=0, storage=0
            )
            ent_subscription.action_confirm()
            ent_subscription.client_order_ref = 'M202303OESUB001'
            ent_subscription._create_invoices()
            ent_subscription.invoice_ids._post()

            # database ping using the Enterprise subscription code
            database = self.create_database()
            self.send_ping(
                database.db_uuid, {'my_awesome_sh_custom': 4200},
                code=ent_subscription.client_order_ref,
                payload_update={'nbr_users': 10, 'nbr_active_users': 5}
            )

        # Link enterprise subscription to Odoo.sh subscription
        odoosh_subscription.odoo_sh_enterprise_subscription_id = ent_subscription

        # PaaS ping using the subscription code set on the project
        self.env.cr.flush()
        with self._mock_saas_config(from_role='paas_master'):
            resp = self.url_open(
                '/odoo-enterprise/paas-subscription-ping',
                auth=SaasAuth(),
                json={'params': {
                    'fetch_only': True,
                    'subscriptions': {
                        odoosh_subscription.client_order_ref: self._sh_repo_ping_info(),
                    }},
                },
            )
            result = resp.json().get('result')

            (odoosh_subscription | ent_subscription).invalidate_recordset()  # ensure cache is clean
            self.assertIn(odoosh_subscription.client_order_ref, result['valid'])
            self.assertEqual(
                result[odoosh_subscription.client_order_ref],
                {
                    'id': odoosh_subscription.id,
                    'enterprise_is_partnership': False,
                    'customer_email': 'enterprise-customer@example.net',
                    'partner_email': False,
                    'users': 10.0,
                    'periodicity': {'unit': 'year', 'duration': 1},
                    'next_invoice_date': '2024-01-15',
                    # values are the one on the sub before the paas ping
                    'workers': 1.0,
                    'staging': 1.0,
                    'storage': 1.0,
                    'dedicated': 1,
                    'enterprise_subscription': {
                        'id': ent_subscription.id,
                        'code': 'M202303OESUB001',
                        'periodicity': {'unit': 'month', 'duration': 1},
                        'next_invoice_date': '2023-02-20',
                    }
                }
            )

    def test_paas_ping_separated_monthly_upsell(self):
        """Test paas ping in case enterprise sub is separated from Odoo.sh sub"""
        yearly_template = 'openerp_enterprise.custom_template_odoo_yearly'
        monthly_template = 'openerp_enterprise.custom_template_odoo_monthly'

        # Create an Odoo.sh subscription (without Odoo users)
        odoosh_subscription = self._create_sh_subscription(template=monthly_template, users=0, workers=1)
        odoosh_subscription.client_order_ref = 'M202303SHSUB001'
        odoosh_subscription.action_confirm()
        odoosh_subscription._create_invoices()
        odoosh_subscription.invoice_ids._post()

        ent_subscription = self._create_sh_subscription(
            template=yearly_template, users=10, workers=0, stagings=0, storage=0
        )
        ent_subscription.action_confirm()
        ent_subscription.client_order_ref = 'M202303OESUB001'
        ent_subscription._create_invoices()
        ent_subscription.invoice_ids._post()

        # Link enterprise subscription to Odoo.sh subscription
        odoosh_subscription.odoo_sh_enterprise_subscription_id = ent_subscription
        odoosh_subscription.odoo_sh_extra_gb = 100
        odoosh_subscription.odoo_sh_extra_stagings = 3
        odoosh_subscription.odoo_sh_extra_workers = 4
        odoosh_subscription.odoo_sh_extra_dedicated = 1
        odoosh_subscription.set_apps_recurring_lines()
        for product, qty in {
            self.env.ref('openerp_enterprise.product_odoo_sh_extra_worker'): 4,
            self.env.ref('openerp_enterprise.product_odoo_sh_extra_gb'): 100,
            self.env.ref('openerp_enterprise.product_odoo_sh_extra_staging'): 3,
            self.env.ref('openerp_enterprise.product_odoo_sh_dedicated'): 1,
        }.items():
            product_line = odoosh_subscription.order_line.filtered(lambda l, p=product: l.product_id == p)
            self.assertEqual(
                product_line.product_uom_qty,
                qty,
                f"The quantity of {product_line.product_id.name} should be {qty}"
            )

    @users('salesman_subscription')
    def test_odoosh_subscription_restrictions(self):
        # adding double product should be forbidden
        subscription = self._create_sh_subscription(workers=2, stagings=4, storage=30)
        with self.assertRaisesRegex(UserError, ".*Odoo SH product.*only have a single line.*"):
            with Form(subscription) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = self.env.ref("openerp_enterprise.product_odoo_sh_extra_worker")
                    new_line.product_uom_qty = 1

        # shared hosting: test for limits forcing use of dedicated hosting
        # - min workers (dedicated required)
        with self.assertRaisesRegex(UserError, "You must use a dedicated Odoo SH Server.*more than.*workers"):
            self._create_sh_subscription(dedicated=0, workers=8 + 1)
        # - min storage (dedicated required)
        with self.assertRaisesRegex(UserError, "You must use a dedicated Odoo SH Server.*more than.*GB"):
            self._create_sh_subscription(dedicated=0, storage=512 + 1)

        # - check worker and storage are depending on each other
        with self.assertRaisesRegex(UserError, "You should have storage.*use Odoo.sh Workers"):
            self._create_sh_subscription(workers=1, storage=0)
        with self.assertRaisesRegex(UserError, "You should have.*Workers.*use Odoo.sh Storage"):
            self._create_sh_subscription(workers=0, storage=1)

        # dedicated hosting: test for lone product or product limits
        # - no lone product
        with self.assertRaisesRegex(UserError, ".*You must have another SH product.*"):
            self._create_sh_subscription(workers=0, stagings=0, storage=0, dedicated=1)
        # - dedicated qty (max. of 1)
        with self.assertRaisesRegex(UserError, ".*ou can't have more than 1 "):
            self._create_sh_subscription(dedicated=2)
        # - workers limit
        with self.assertRaisesRegex(UserError, "A dedicated Odoo SH.*more than.*workers"):
            self._create_sh_subscription(dedicated=1, workers=64 + 1)
        # - storage limit
        with self.assertRaisesRegex(UserError, "A dedicated Odoo SH.*more than.*GB"):
            self._create_sh_subscription(dedicated=1, storage=4096 + 1)

    @users('salesman_subscription')
    def test_odoosh_subscription_separate_enterprise_sub(self):
        # separate enterprise subscription: could not have odoo user on both subscriptions
        sh_subscription = self._create_sh_subscription(workers=1, users=0)
        ent_subscription = self._create_sh_subscription(users=10, workers=0, stagings=0, storage=0)
        sh_subscription.odoo_sh_enterprise_subscription_id = ent_subscription

        # Ensure that we can't add users to odoo.sh sub if it's linked to a
        # separate enterprise subscription
        with self.assertRaisesRegex(
            ValidationError, "You can't.*Odoo.sh Enterprise subscription.*Odoo Users.*same subscription"
        ):
            with Form(sh_subscription) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = self.env.ref('openerp_enterprise.product_odoo_user_custom')
                    new_line.product_uom_qty = 24

    @users('salesman_subscription')
    def test_renewal_move_info(self):
        self.env.user.email = "salesman@odoo.com"
        ent_subscription = self._create_sh_subscription(users=10, workers=0, stagings=0, storage=0)
        sh_subscription = self._create_sh_subscription(workers=1, users=0)
        sh_subscription.odoo_sh_enterprise_subscription_id = ent_subscription

        ent_subscription.action_confirm()
        sh_subscription.action_confirm()

        ent_subscription._create_invoices().action_post()
        sh_subscription._create_invoices().action_post()

        self.send_ping(
            generate_uuid(), {},
            code=ent_subscription.client_order_ref,
            payload_update={'nbr_users': 10, 'nbr_active_users': 5}
        )
        ent_subscription.is_staff_restricted = True

        database = ent_subscription.database_ids

        project = self.env['project.project'].sudo().create({'name': 'Test project'})
        task = self.env['project.task'].sudo().create({
            'name': 'Test Task',
            'mnt_subscription_id': ent_subscription.id,
            'project_id': project.id
        })

        gh_user = self.env['openerp.enterprise.github.user'].sudo().create({
            'subscription_id': ent_subscription.id,
            'name': 'gh_test_user',
        })
        self.assertEqual(task.mnt_subscription_id, ent_subscription)
        self.assertEqual(sh_subscription.odoo_sh_enterprise_subscription_id, ent_subscription)
        action = ent_subscription.prepare_renewal_order()
        new_ent_sub = self.env['sale.order'].browse(action['res_id'])
        first_year_disc_prod = self.env.ref('openerp_enterprise.product_odoo_user_custom_discount')
        new_ent_sub.order_line.filtered(lambda l: l.product_id == first_year_disc_prod).unlink()
        new_ent_sub.action_confirm()
        self.cr.precommit.run()
        self.assertTrue(new_ent_sub.is_staff_restricted, "Staff restriction should propagate to renewal")
        self.assertEqual(task.mnt_subscription_id, new_ent_sub, "Task should be moved at renewal")
        self.assertEqual(
            sh_subscription.odoo_sh_enterprise_subscription_id,
            new_ent_sub,
            "Odoo sh subscription should be moved at renewal"
        )
        self.assertEqual(new_ent_sub.database_ids, database, "Database should be moved at renewal")
        self.assertEqual(new_ent_sub.github_user_ids, gh_user, "Github users should be moved at renewal")

        action = sh_subscription.prepare_renewal_order()
        new_sh_sub = self.env['sale.order'].browse(action['res_id'])
        new_sh_sub.action_confirm()
        self.cr.precommit.run()
        self.assertEqual(
            new_sh_sub.odoo_sh_enterprise_subscription_id,
            new_ent_sub,
            "Main subscription should be moved at renewal"
        )

    def test_renewal_from_db(self):
        """Test flow when user click on 'Renew' from its database expiration panel."""
        self.main_subscription._create_invoices()._post()
        self.assertEqual(self.main_subscription.subscription_state, '3_progress')

        db_uuid = generate_uuid()
        res = self.send_ping(db_uuid, {}, payload_update={
            'enterprise_code': self.main_subscription.client_order_ref,
        })
        self.env['openerp.enterprise.database'].search([('db_uuid', '=', db_uuid)])
        self.assertEqual(res['enterprise_info']['expiration_reason'], 'renewal')

        # Create a renewal
        renewal_order = self.renew(self.main_subscription)
        self.assertEqual(renewal_order.state, 'draft')
        self.assertEqual(renewal_order.subscription_state, '2_renewal')

        # Then create an upsell
        # (guard, last created order with the same subscription code => we should never redirect it)
        upsell_order = self.upsell(self.main_subscription)
        self.assertEqual(upsell_order.state, 'draft')
        self.assertEqual(upsell_order.subscription_state, '7_upsell')

        self.authenticate(self.customer_user.email, self.customer_user.email)

        # Subscription still valid, Renewal not sent yet => we should redirect to current sub.
        resp = self.url_open(
            f"/odoo-enterprise/renew?contract={self.main_subscription.client_order_ref}",
            allow_redirects=False
        )
        self.assertEqual(
            resp.headers['Location'],
            f"{self.base_url()}/my/subscriptions/{self.main_subscription.id}",
            "We should redirect user to the most valid subscription with provided ent. code",
        )

        # Subscription still valid, but a "sent" renewal exists => redirect to the renewal
        renewal_order.action_quotation_sent()
        resp = self.url_open(
            f"/odoo-enterprise/renew?contract={self.main_subscription.client_order_ref}",
            allow_redirects=False
        )
        self.assertEqual(
            resp.headers['Location'],
            f"{self.base_url()}/my/orders/{renewal_order.id}",
            "We should redirect user to the last send renewal with provided ent. code",
        )

    def test_welcome_email_sent(self):
        self.assertTrue(
            self.main_subscription.message_ids.filtered_domain(
                [('subject', '=like', 'Your subscription % is active!')]
            ),
            "Welcome email should have been sent"
        )
        self.main_subscription._create_invoices()._post()
        upsell_order = self.upsell(self.main_subscription)
        upsell_order.action_confirm()
        self.assertFalse(
            len(upsell_order.message_ids.filtered_domain(
                [('subject', '=like', 'Your subscription % is active!')]
            )),
            "Welcome shouldn't be sent on upsell"
        )
        renewal_order = self.renew(self.main_subscription)
        renewal_order.action_confirm()
        self.assertFalse(
            renewal_order.message_ids.filtered_domain(
                [('subject', '=like', 'Your subscription % is active!')]
            ),
            "Welcome shouldn't be sent at renewal"
        )

    def test_welcome_odoosh_email_sent(self):
        sh_dedicated_subs = self._create_sh_subscription(workers=8, stagings=1, storage=100, dedicated=1, users=1)
        sh_dedicated_subs.action_confirm()
        sh_subs = self._create_sh_subscription(workers=2, stagings=1, storage=100, dedicated=0, users=1)
        sh_subs.action_confirm()
        self.cr.precommit.run()

        self.assertTrue(
            sh_dedicated_subs.message_ids.filtered_domain(
                [('subject', '=like', 'Subscription %: Attention Needed.')]
            ),
            "Sh subscription with dedicated server should trigger an email to odoo.sh team"
        )
        self.assertTrue(
            sh_dedicated_subs.message_ids.filtered_domain(
                [('subject', '=', 'Welcome to Odoo.sh!')]
            ),
            "Sh subscription confirmation should send a welcome email"
        )

        self.assertTrue(
            sh_subs.message_ids.filtered_domain(
                [('subject', '=', 'Welcome to Odoo.sh!')]
            ),
            "Sh subscription confirmation should send a welcome email"
        )


@tagged("post_install", "-at_install")
class TestHosting(SaasWorkerHttpCommon, SubscriptionCommon):
    std_quotation_template = SubscriptionCommon.qt_standard_monthly
    cust_quotation_template = SubscriptionCommon.qt_custom_monthly
    quotation_template = SubscriptionCommon.qt_custom_monthly

    @classmethod
    def init_master_data(cls):
        super().init_master_data()
        cls.customer_user = cls.env['res.users'].create({
            'login': cls.customer.email,
            'password': cls.customer.email,
            'partner_id': cls.customer.id,
            'groups_id': [(6, 0, cls.env.ref('base.group_portal').ids)],
        })

    def test_subscription_backend_flow(self):
        # Create a new Subscription Quotation
        sub1 = self._create_so_from_template(self.customer, self.std_quotation_template)
        self.assertEqual(sub1.hosting, "saas", "Default hosting should be Online (saas)")

        # Ensure hosting is enforced to "saas" for Standard Plan
        sub1.hosting = "premise"
        self.assertEqual(sub1.hosting, "saas", "For Standard Plan, enforce hosting to Online (saas)")

        # Changing plan to custom, we should now allow all hosting types
        with Form(sub1) as sub1_form:
            sub1_form.sale_order_template_id = self.env.ref(self.cust_quotation_template)
        self.assertEqual(sub1.hosting, "saas")
        sub1.hosting = "premise"
        self.assertEqual(sub1.hosting, "premise")

        # Confirm SO and invoice it
        sub1.action_confirm()
        invoice1 = sub1._create_invoices()
        self.assertTrue(
            all(not line.hosting for line in invoice1.line_ids),
            "We should not set/force hosting when creating draft invoice",
        )
        self.assertEqual(
            invoice1.hosting, 'premise',
            "We should still correctly auto-compute the hosting if not value are forced"
        )
        invoice1._post()
        self.assertTrue(
            all(line.hosting == "premise" for line in invoice1.invoice_line_ids),
            "Invoice hosting should be the same as Subscription when posting an invoice",
        )

        # Customer create a database and register it
        db_uuid = generate_uuid()
        # Force ping as "premise"
        with patch("odoo.addons.openerp_enterprise.models.openerp_enterprise.get_roles_from_ip",
                   return_value={}):
            self.send_ping(db_uuid, {})
        db = self.env["openerp.enterprise.database"].search([('db_uuid', '=', db_uuid)])
        self.assertEqual(db.hosting, "premise")
        self.assertEqual(sub1.hosting, "premise")

        # In the meantime customer try out SaaS, so it create his database there and it
        # and register his subscription (subscription should switch to "SaaS" hosting)
        db2_uuid = generate_uuid()
        self.send_ping(db2_uuid, {}, code=sub1.client_order_ref)
        db2 = self.env["openerp.enterprise.database"].search([('db_uuid', '=', db2_uuid)])
        self.assertEqual(db2.hosting, "saas")
        self.assertEqual(db2.subscription_id, sub1)
        # db2.subscription_id = sub1
        self.assertEqual(sub1.hosting, "saas")

        # Old database still ping, ensure the hosting stay as "SaaS"
        with patch("odoo.addons.openerp_enterprise.models.openerp_enterprise.get_roles_from_ip",
                   return_value={}):
            self.send_ping(db_uuid, {})
        self.assertEqual(sub1.hosting, "saas")

        # Few times has passed, the customer now need more custom., so the
        # salesman create an upsell for Odoo.SH
        action = sub1.prepare_upsell_order()
        upsell_order = self.env["sale.order"].browse(action["res_id"])
        self.assertEqual(upsell_order.hosting, "saas")
        with Form(upsell_order) as upsell_form:
            with upsell_form.order_line.new() as new_line:
                new_line.product_id = self.env.ref("openerp_enterprise.product_odoo_sh_extra_worker")
                new_line.product_uom_qty = 1
        self.assertEqual(upsell_order.hosting, "paas", "Adding Odoo.SH should force hosting to Odoo.SH (paas)")
        upsell_order.action_confirm()
        self.assertEqual(sub1.hosting, "paas", "Parent Subscription hosting should also be Odoo.SH (paas)")
        invoice2 = upsell_order._create_invoices()
        invoice2._post()
        self.assertTrue(
            all(line.hosting == "paas" for line in invoice2.invoice_line_ids),
            "Precedingly posted invoice should retain the previous hosting"
        )
        invoice1.invalidate_recordset([])
        self.assertTrue(
            all(line.hosting == "premise" for line in invoice1.invoice_line_ids),
            "Precedingly posted invoice should retain the previous hosting"
        )

        # There was a mistake on the original invoice, refund it & create a new invoice
        # using the reversal wizard
        move_reversal = self.env["account.move.reversal"]\
            .with_context(active_model="account.move", active_ids=invoice1.ids)\
            .create({
                "journal_id": invoice1.journal_id.id,
                "reason": "no reason",
            })
        reversal = move_reversal.modify_moves()
        invoice1_new_invoice = self.env["account.move"].browse(reversal["res_id"])

        # there should credit note for invoice1, and it should have the
        # same hosting as invoice1
        self.assertTrue(invoice1.reversal_move_id)
        invoice1_credit_note = invoice1.reversal_move_id
        self.assertEqual(invoice1_credit_note.hosting, "premise")
        self.assertTrue(all(line.hosting == "premise" for line in invoice1_credit_note.invoice_line_ids))

        # same for new invoice, in that case hosting must be already forced
        # on the invoice lines by the reversal wizard
        self.assertEqual(invoice1_new_invoice.hosting, "premise")
        self.assertTrue(all(line.hosting == "premise" for line in invoice1_new_invoice.invoice_line_ids))

        # but in case the user manually copy an invoice, lines linked to a
        # subscription should retain the 'subscription_id', so the hosting
        # should stay the same
        invoice1_copy = invoice1.copy({})
        self.assertTrue(invoice1_copy.state, "draft")
        self.assertTrue(all(not line.hosting for line in invoice1_copy.invoice_line_ids))
        self.assertEqual(invoice1_copy.hosting, 'paas')

    def test_renewal_default_hosting(self):
        """Ensure that upon renewal, the hosting take the same value as the parent subscription."""
        sub1 = self._create_so_from_template(self.customer, self.cust_quotation_template)
        sub1.hosting = 'premise'
        sub1.action_confirm()
        sub1._create_invoices()._post()

        sub1_renewal = self.renew(sub1)
        self.assertEqual(sub1_renewal.hosting, 'premise')

    def test_renewal_change_of_hosting(self):
        """Ensure that on renewal, the user can change the hosting."""
        sub1 = self._create_so_from_template(self.customer, self.std_quotation_template)
        sub1.hosting = 'saas'
        sub1.action_confirm()
        sub1._create_invoices()._post()

        sub1_renewal = self.renew(sub1)
        # trying setting paas without changing the plan shoul   d revert to saas
        with Form(sub1_renewal) as sub1_renewal_form:
            sub1_renewal_form.hosting = 'paas'
        self.assertEqual(sub1_renewal.hosting, 'saas')

        # change the plan, we can now change the hosting to any hosting
        with Form(sub1_renewal) as sub1_renewal_form:
            sub1_renewal_form.sale_order_template_id = self.env.ref(self.cust_quotation_template)
            sub1_renewal_form.hosting = 'paas'
        self.assertEqual(
            sub1_renewal_form.hosting, 'paas',
            "On renewal the change of hosting should be permitted"
        )

    def test_move_line_hosting_inverse(self):
        """Test that changing move line 'hosting' correctly update move's hosting."""
        so_form = Form(self.env["sale.order"])
        so_form.partner_id = self.customer
        so_form.sale_order_template_id = self.env.ref(self.cust_quotation_template)
        so_form.hosting = 'saas'
        so = so_form.save()
        so.action_confirm()

        # Create draft invoice
        draft_invoice = so._create_invoices()
        self.assertEqual(draft_invoice.hosting, 'saas')
        so_mnt_line = so._get_maintenance_line()
        draft_invoice_mnt_line = so_mnt_line.invoice_lines & draft_invoice.invoice_line_ids
        # Setting Maintenance Line to 'premise' hosting should give mixed lines
        # result so that move's should be set to False (not determined)
        draft_invoice_mnt_line.hosting = 'premise'
        self.assertEqual(draft_invoice.hosting, False)
        # Setting remaining lines to 'premise' as well, the move's hosting
        # should now also be 'premise'
        (draft_invoice.invoice_line_ids - draft_invoice_mnt_line).hosting = 'premise'
        self.assertEqual(draft_invoice.hosting, 'premise')

    def test_referrer_force_manual_hosting_selection(self):
        so_form = Form(self.env["sale.order"])
        so_form.partner_id = self.customer
        so_form.sale_order_template_id = self.env.ref(self.cust_quotation_template)
        self.assertEqual(so_form.hosting, "saas")
        so_form.referrer_id = self.env.ref("base.partner_admin")
        self.assertFalse(so_form.hosting)
        with self.assertRaisesRegex(AssertionError, "hosting is a required field"):
            so_form.save()
        so_form.hosting = 'premise'
        so = so_form.save()
        self.assertEqual(so.hosting, "premise")

    def test_self_oe(self):
        # When user create a subscription from it's database ("Upgrade",
        # "Buy" for db in trial-15days period) we should ensure that we take
        # the same hosting as the database the user request from.
        hosting_to_roles = {'premise': set(), 'saas': {'trial'}, 'paas': {'paas_worker'}}
        self.authenticate(self.customer.email, self.customer.email)
        for expected_hosting, saas_worker_roles in hosting_to_roles.items():
            with self.subTest(hosting=expected_hosting):
                db_uuid = generate_uuid()
                with patch("odoo.addons.openerp_enterprise.models.openerp_enterprise.get_roles_from_ip",
                           return_value=saas_worker_roles):
                    self.send_ping(db_uuid, {})
                db = self.env["openerp.enterprise.database"].search([('db_uuid', '=', db_uuid)])
                self.assertEqual(db.hosting, expected_hosting, "DB hosting does not match")

                resp = self.url_open('/odoo-enterprise/upgrade-online/%s' % db_uuid, allow_redirects=False)
                resp.raise_for_status()
                db_sub1 = db.subscription_id
                self.assertEqual(db_sub1.hosting, expected_hosting)

    @users('salesman_subscription')
    def test_odoo_2022_pricing_rules_sub_duplicated_lines(self):

        # Pricing 2022 Products
        user_standard = self.env.ref('openerp_enterprise.product_odoo_user_standard')
        user_standard_discount = self.env.ref('openerp_enterprise.product_odoo_user_standard_discount')
        user_custom = self.env.ref('openerp_enterprise.product_odoo_user_custom')
        user_custom_discount = self.env.ref('openerp_enterprise.product_odoo_user_custom_discount')
        once_discount = self.env.ref('openerp_enterprise.product_odoo_once_discount')
        recu_discount = self.env.ref('openerp_enterprise.product_odoo_recu_discount')

        # Excluded from double line check
        subscription_extra_discount = self.env.ref('openerp_enterprise.product_odoo_subscription_extra_discount')
        subscription_referral_per_country = self.env.ref('openerp_enterprise.product_odoo_subscription_referral_per_country')
        subscription_monthly_success_pack_basic = self.env.ref('openerp_enterprise.product_odoo_subscription_monthly_success_pack_basic')
        subscription_monthly_success_pack_standard = self.env.ref('openerp_enterprise.product_odoo_subscription_monthly_success_pack_standard')
        learning_partner_discount = self.env.ref('openerp_enterprise.product_learning_partner_discount')
        official_partner_discount = self.env.ref('openerp_enterprise.product_official_partner_discount')

        subscription = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        with self.assertRaisesRegex(UserError, ".*Each Pricing Plan product line should be unique.*"):
            with Form(subscription) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = user_standard
                    new_line.product_uom_qty = 1

        subscription = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        with self.assertRaisesRegex(UserError, ".*Each Pricing Plan product line should be unique.*"):
            with Form(subscription) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = user_standard_discount
                    new_line.product_uom_qty = 1

        subscription = self._create_so_from_template(self.customer, SubscriptionCommon.qt_custom_monthly)
        with self.assertRaisesRegex(UserError, ".*Each Pricing Plan product line should be unique.*"):
            with Form(subscription) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = user_custom
                    new_line.product_uom_qty = 1

        subscription = self._create_so_from_template(self.customer, SubscriptionCommon.qt_custom_monthly)
        with self.assertRaisesRegex(UserError, ".*Each Pricing Plan product line should be unique.*"):
            with Form(subscription) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = user_custom_discount
                    new_line.product_uom_qty = 1

        # Should not raise anything
        subscription = self._create_so_from_template(self.customer, SubscriptionCommon.qt_custom_monthly)
        with sub_form.order_line.new() as new_line:
            new_line.product_id = once_discount
            new_line.product_uom_qty = 1
        with sub_form.order_line.new() as new_line:
            new_line.product_id = once_discount
            new_line.product_uom_qty = 1

        subscription = self._create_so_from_template(self.customer, SubscriptionCommon.qt_custom_monthly)
        with self.assertRaisesRegex(UserError, ".*Each Pricing Plan product line should be unique.*"):
            with Form(subscription) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = recu_discount
                    new_line.product_uom_qty = 1
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = recu_discount
                    new_line.product_uom_qty = 1

        # Should not raise anything
        subscription = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        with Form(subscription) as sub_form:
            with sub_form.order_line.new() as new_line:
                new_line.product_id = subscription_extra_discount
                new_line.product_uom_qty = 1
            with sub_form.order_line.new() as new_line:
                new_line.product_id = subscription_extra_discount
                new_line.product_uom_qty = 1
            with sub_form.order_line.new() as new_line:
                new_line.product_id = subscription_referral_per_country
                new_line.product_uom_qty = 1
            with sub_form.order_line.new() as new_line:
                new_line.product_id = subscription_referral_per_country
                new_line.product_uom_qty = 1
            with sub_form.order_line.new() as new_line:
                new_line.product_id = subscription_monthly_success_pack_basic
                new_line.product_uom_qty = 1
            with sub_form.order_line.new() as new_line:
                new_line.product_id = subscription_monthly_success_pack_basic
                new_line.product_uom_qty = 1
            with sub_form.order_line.new() as new_line:
                new_line.product_id = subscription_monthly_success_pack_standard
                new_line.product_uom_qty = 1
            with sub_form.order_line.new() as new_line:
                new_line.product_id = subscription_monthly_success_pack_standard
                new_line.product_uom_qty = 1
            with sub_form.order_line.new() as new_line:
                new_line.product_id = learning_partner_discount
                new_line.product_uom_qty = 1
            with sub_form.order_line.new() as new_line:
                new_line.product_id = learning_partner_discount
                new_line.product_uom_qty = 1
            with sub_form.order_line.new() as new_line:
                new_line.product_id = official_partner_discount
                new_line.product_uom_qty = 1
            with sub_form.order_line.new() as new_line:
                new_line.product_id = official_partner_discount
                new_line.product_uom_qty = 1

    @users('salesman_subscription')
    def test_odoo_2022_pricing_rules_sub_plan_mismatch(self):

        # Plan products
        user_standard = self.env.ref('openerp_enterprise.product_odoo_user_standard')
        user_standard_discount = self.env.ref('openerp_enterprise.product_odoo_user_standard_discount')
        user_custom = self.env.ref('openerp_enterprise.product_odoo_user_custom')
        user_custom_discount = self.env.ref('openerp_enterprise.product_odoo_user_custom_discount')

        subscription = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        regex = r".*The product\(s\) \(Custom Plan\) cannot be added to a standard plan.*"
        with self.assertRaisesRegex(UserError, regex):
            with Form(subscription) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = user_custom
                    new_line.product_uom_qty = 1

        subscription = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        regex = r".*The product\(s\) \(Custom Plan First Year Discount\) cannot be added to a standard plan.*"
        with self.assertRaisesRegex(UserError, regex):
            with Form(subscription) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = user_custom_discount
                    new_line.product_uom_qty = 1

        subscription = self._create_so_from_template(self.customer, SubscriptionCommon.qt_custom_monthly)
        regex = r".*The product\(s\) \(Standard Plan\) cannot be added to a custom plan.*"
        with self.assertRaisesRegex(UserError, regex):
            with Form(subscription) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = user_standard
                    new_line.product_uom_qty = 1

        subscription = self._create_so_from_template(self.customer, SubscriptionCommon.qt_custom_monthly)
        regex = r".*The product\(s\) \(Standard Plan First Year Discount\) cannot be added to a custom plan.*"
        with self.assertRaisesRegex(UserError, regex):
            with Form(subscription) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = user_standard_discount
                    new_line.product_uom_qty = 1

    @users('salesman_subscription')
    def test_odoo_2022_pricing_rules_sub_plan_recurrence_mismatch(self):

        plan_yearly = self.env.ref('sale_subscription.subscription_plan_year')
        plan_2years = self.env.ref('openerp_enterprise.subscription_plan_odoo_2_years_plan')

        sub = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        error_regex_recurrence_mismatch = r".*Please change the plan or the recurrence to save the quote.*"

        # Standard Recurrence
        with self.assertRaisesRegex(UserError, error_regex_recurrence_mismatch):
            sub.plan_id = plan_yearly

        # Internal Recurrence
        with self.assertRaisesRegex(UserError, error_regex_recurrence_mismatch):
            sub.plan_id = plan_2years

    @users('salesman_subscription')
    def test_odoo_2022_pricing_rules_archived_product(self):

        sub = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        error_regex_archived_product = r".*You cannot confirm a quote with archived products.*"

        archived_product = self.env['product.product'].sudo().create({
            'name': 'Archived Product',
            'categ_id': self.env.ref('product.product_category_all').id,
            'type': 'service',
            'active': False,
        })

        with self.assertRaisesRegex(UserError, error_regex_archived_product):
            with Form(sub) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = archived_product
                    new_line.product_uom_qty

    @users('salesman_subscription')
    def test_odoo_2022_pricing_rules_upsell_no_recurring_products(self):

        sub = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        sub.action_confirm()
        sub._create_invoices()._post()
        action = sub.prepare_upsell_order()
        upsell_so = self.env['sale.order'].browse(action['res_id'])

        # Shouldn't raise anything
        with Form(upsell_so) as upsell_form:
            with upsell_form.order_line.new() as new_line:
                new_line.product_id = self.env.ref('openerp_enterprise.product_odoo_once_discount')
                new_line.product_uom_qty = 1

        with self.assertRaisesRegex(UserError, r".*You can only upsell recurring products.*"):
            with Form(upsell_so) as upsell_form:
                with upsell_form.order_line.new() as new_line:
                    new_line.product_id = self.certification_product
                    new_line.product_uom_qty = 1

        with self.assertRaisesRegex(UserError, r".*You can only upsell recurring products.*"):
            with Form(upsell_so) as upsell_form:
                with upsell_form.order_line.new() as new_line:
                    new_line.product_id = self.success_pack_product
                    new_line.product_uom_qty = 1

    @users('salesman_subscription')
    def test_odoo_2022_pricing_rules_commison_no_discount_products(self):

        zero_percent_comm = self.env.ref('openerp_portal.zero_percent_commission_to_partner')
        ready_2023_comm = self.env.ref('openerp_portal.ready_commission_plan_2023')

        sub = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        sub.commission_plan_id = zero_percent_comm

        discount_lp_tmpl = sub.env.ref('openerp_enterprise.product_learning_partner_discount_product_template')
        discount_op_tmpl = sub.env.ref('openerp_enterprise.product_official_partner_discount_product_template')

        # Shouldn't raise anything
        with Form(sub) as sub_form:
            with sub_form.order_line.new() as new_line:
                new_line.product_id = discount_lp_tmpl.product_variant_id
                new_line.product_uom_qty = 1

        sub.commission_plan_id = ready_2023_comm
        with Form(sub) as sub_form:
            with sub_form.order_line.new() as new_line:
                new_line.product_id = discount_op_tmpl.product_variant_id
                new_line.product_uom_qty = 1

        # Assert that the 0% commission plan is enforced if the referrer is the sub's partner
        sub = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        sub.commission_plan_id = ready_2023_comm
        sub.referrer_id = sub.partner_id

        regex = r".*You cannot have a partner discount product unless the subscription has a 0%% commission plan.*"
        with self.assertRaisesRegex(UserError, regex):
            with Form(sub) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = discount_op_tmpl.product_variant_id
                    new_line.product_uom_qty = 1

    @users('salesman_subscription')
    def test_odoo_bypass_rule_on_internal_note(self):
        sub = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        sub.action_confirm()
        sub._create_invoices()._post()
        action = sub.prepare_renewal_order()
        child_sub = self.env['sale.order'].browse(action['res_id'])
        child_sub.action_confirm()
        custom_template = self.env.ref(SubscriptionCommon.qt_custom_monthly)
        # Make the parent wrong
        sub.sudo().write({'sale_order_template_id': custom_template})
        # Should be able to write on internal_note
        child_sub.write({'internal_note_display': 'saas'})
        # With other fields as well
        child_sub.write({'internal_note_display': 'saas', 'hosting': 'saas'})
        # Should be able to write on internal_note
        sub.write({'internal_note_display': 'saas'})
        # Should not be able to write on parent sub other fields
        regex = r".*The product\(s\) \(Standard Plan, Standard Plan First Year Discount\) cannot be added to a custom plan.*"
        with self.assertRaisesRegex(UserError, regex):
            sub.write({'internal_note_display': 'saas', 'hosting': 'saas'})

    @users('salesman_subscription')
    def test_odoo_rule_valid_sub_origin(self):

        regex = r".*is either not in a valid state or never started.*"

        # force no Start date
        sub = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        sub.action_confirm()
        sub._create_invoices()._post()
        with self.assertRaisesRegex(UserError, regex):
            sub.start_date = None
            sub.prepare_upsell_order()

        sub = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        sub.action_confirm()
        sub._create_invoices()._post()
        with self.assertRaisesRegex(UserError, regex):
            sub.start_date = None
            sub.prepare_renewal_order()

    @users('salesman_subscription')
    def test_odoo_2022_pricing_rules_no_lone_maintenance_line(self):

        product_maintenance = self.env.ref('openerp_enterprise.product_maintenance')

        sub = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_monthly)
        regex = r".*a custom plan or a separate main subscription on which the maintenance applies.*"
        with self.assertRaisesRegex(UserError, regex):
            with Form(sub) as sub_form:
                with sub_form.order_line.new() as new_line:
                    new_line.product_id = product_maintenance
                    new_line.product_uom_qty = 1

    @users('salesman_subscription')
    def test_odoo_2022_pricing_rules_no_lone_maintenance_line_switch_plan(self):
        sub = self._create_so_from_template(self.customer, SubscriptionCommon.qt_standard_yearly)
        sub.action_confirm()
        sub._create_invoices()._post()
        action = sub.prepare_upsell_order()
        upsell_so = self.env['sale.order'].browse(action['res_id'])
        odoo_user_standard = self.env.ref('openerp_enterprise.product_odoo_user_standard')
        odoo_user_custom = self.env.ref('openerp_enterprise.product_odoo_user_custom')
        # make sure it's a switching plan upsell order
        self.assertTrue(upsell_so.order_line.filtered(lambda l: l.product_id == odoo_user_standard))
        self.assertTrue(upsell_so.order_line.filtered(lambda l: l.product_id == odoo_user_custom))
        upsell_so.action_confirm()
        self.assertEqual(upsell_so.state, 'sale')
